/* 

        Copyright (C) 1995-
        Free Software Foundation, Inc.

   This file is part of GNU cfengine - written and maintained 
   by Mark Burgess, Dept of Computing and Engineering, Oslo College,
   Dept. of Theoretical physics, University of Oslo
 
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

*/

/*****************************************************************************/
/*                                                                           */
/* File: ip.c                                                                */
/*                                                                           */
/* Created: Sun Feb 17 13:47:19 2002                                         */
/*                                                                           */
/*****************************************************************************/

#include "cf.defs.h"
#include "cf.extern.h"

/*****************************************************************************/

int RemoteConnect(host,forceipv4)          /* Handle ipv4 or ipv6 connection */

char *host, forceipv4;

{ int err;

#ifdef HAVE_GETADDRINFO /* --------------we have ipv6 ------------------- */
 
if (forceipv4 == 'n')
   {
   struct addrinfo query, *response, *ap;
   struct addrinfo query2, *response2, *ap2;
   int err,connected = false;
   
   bzero(&query,sizeof(struct addrinfo));   

   query.ai_family = AF_UNSPEC;
   query.ai_socktype = SOCK_STREAM;
   
   if ((err = getaddrinfo(host,"5308",&query,&response)) != 0)
      {
      snprintf(OUTPUT,bufsize,"Unable to lookup hostname or cfengine service: %s",gai_strerror(err));
      CfLog(cferror,OUTPUT,"");
      return false;
      }
   
   for (ap = response; ap != NULL; ap = ap->ai_next)
      {
      Verbose("Connect to %s = %s on port cfengine\n",host,sockaddr_ntop(ap->ai_addr));
      
      if ((CONN->sd = socket(ap->ai_family,ap->ai_socktype,ap->ai_protocol)) == -1)
	 {
	 CfLog(cfinform,"Couldn't open a socket","socket");      
	 continue;
	 }

      if (BINDINTERFACE[0] != '\0')
	 {
	 bzero(&query2,sizeof(struct addrinfo));   

	 query.ai_family = AF_UNSPEC;
	 query.ai_socktype = SOCK_STREAM;
	 
	 if ((err = getaddrinfo(BINDINTERFACE,NULL,&query2,&response2)) != 0)
	    {
	    snprintf(OUTPUT,bufsize,"Unable to lookup hostname or cfengine service: %s",gai_strerror(err));
	    CfLog(cferror,OUTPUT,"");
	    return false;
	    }

	 for (ap2 = response2; ap2 != NULL; ap2 = ap2->ai_next)
	    {
	    if (bind(CONN->sd, ap2->ai_addr, ap2->ai_addrlen) == 0)
	       {
	       freeaddrinfo(response2);
	       break;
	       }
	    }

	 if (response2)
	    {
	    free(response2);
	    }
	 }
      
      signal(SIGALRM,(void *)TimeOut);
      alarm(CF_TIMEOUT);
      
      if (connect(CONN->sd,ap->ai_addr,ap->ai_addrlen) >= 0)
	 {
	 connected = true;
	 alarm(0);
	 signal(SIGALRM,SIG_DFL);
	 break;
	 }
      
      alarm(0);
      signal(SIGALRM,SIG_DFL);
      }
   
   if (connected)
      {
      CONN->family = ap->ai_family;
      snprintf(CONN->remoteip,cfmaxiplen-1,"%s",sockaddr_ntop(ap->ai_addr));
      }
   else
      {
      close(CONN->sd);
      snprintf(OUTPUT,bufsize*2,"Couldn't connect to host %s\n",host);
      CONN->sd = cf_not_connected;
      }

   if (response != NULL)
      {
      freeaddrinfo(response);
      }

   if (!connected)
      {
      return false;
      }
   }
 
 else
     
#endif /* ---------------------- only have ipv4 ---------------------------------*/ 

   {
   struct hostent *hp;
   struct sockaddr_in cin;
   bzero(&cin,sizeof(cin));
   
   if ((hp = gethostbyname(host)) == NULL)
      {
      snprintf(OUTPUT,bufsize,"Unable to look up IP address of %s",host);
      CfLog(cferror,OUTPUT,"gethostbyname");
      return false;
      }
   
   cin.sin_port = CfenginePort();
   cin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
   cin.sin_family = AF_INET; 
   
   Verbose("Connect to %s = %s, port h=%d\n",host,inet_ntoa(cin.sin_addr),htons(PORTNUMBER));
    
   if ((CONN->sd = socket(AF_INET,SOCK_STREAM,0)) == -1)
      {
      CfLog(cferror,"Couldn't open a socket","socket");
      return false;
      }

   if (BINDINTERFACE[0] != '\0')
      {
      Verbose("Cannot bind interface with this OS.\n");
      /* Could fix this - any point? */
      }
   
   CONN->family = AF_INET;
   snprintf(CONN->remoteip,cfmaxiplen-1,"%s",inet_ntoa(cin.sin_addr));
    
   signal(SIGALRM,(void *)TimeOut);
   alarm(CF_TIMEOUT);
    
   if (err=connect(CONN->sd,(void *)&cin,sizeof(cin)) == -1)
      {
      snprintf(OUTPUT,bufsize*2,"Couldn't connect to host %s\n",host);
      CfLog(cfinform,OUTPUT,"connect");
      return false;
      }
   
   alarm(0);
   signal(SIGALRM,SIG_DFL);
   }

LastSeen(host);
return true; 
}



/*******************************************************************/

short CfenginePort()

{
struct servent *server;

if ((server = getservbyname(CFENGINE_SERVICE,"tcp")) == NULL)
   {
   CfLog(cflogonly,"Couldn't get cfengine service, using default","getservbyname");
   return htons(5308);
   }

return server->s_port;
}


/*****************************************************************************/

char *Hostname2IPString(hostname)

char *hostname;

{ static char ipbuffer[65];
  int err;
 
#ifdef HAVE_GETADDRINFO /* --------------we have ipv6 ------------------- */

 struct addrinfo query, *response, *ap;

 bzero(&query,sizeof(struct addrinfo));   
 query.ai_family = AF_UNSPEC;
 query.ai_socktype = SOCK_STREAM;

 bzero(ipbuffer,63);
 
if ((err = getaddrinfo(hostname,NULL,&query,&response)) != 0)
   {
   snprintf(OUTPUT,bufsize,"Unable to lookup hostname (%s) or cfengine service: %s",hostname,gai_strerror(err));
   CfLog(cferror,OUTPUT,"");
   return hostname;
   }
 
for (ap = response; ap != NULL; ap = ap->ai_next)
   {
   strncpy(ipbuffer,sockaddr_ntop(ap->ai_addr),64);
   Debug("Found address (%s) for host %s\n",ipbuffer,hostname);
   
   freeaddrinfo(response);   
   return ipbuffer;
   }
#else
 struct hostent *hp;
 struct sockaddr_in cin;
 bzero(&cin,sizeof(cin));

 bzero(ipbuffer,63);

if ((hp = gethostbyname(hostname)) != NULL)
   {
   cin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
   strncpy(ipbuffer,inet_ntoa(cin.sin_addr),63);
   Verbose("Found address (%s) for host %s\n",ipbuffer,hostname);
   return ipbuffer;
   }
#endif
   
snprintf(ipbuffer,63,"Unknown IP %s",hostname);
return ipbuffer;
}


/*****************************************************************************/

char *IPString2Hostname(ipaddress)

char *ipaddress; /* String form */

{ static char hostbuffer[128];
  int err;

#ifdef HAVE_GETADDRINFO

 struct addrinfo query, *response, *ap;

bzero(&query,sizeof(query));
bzero(&response,sizeof(response));

query.ai_flags = AI_CANONNAME;

bzero(hostbuffer,128);

if ((err = getaddrinfo(ipaddress,NULL,&query,&response)) != 0)
   {
   snprintf(OUTPUT,bufsize,"Unable to lookup IP address (%s): %s",ipaddress,gai_strerror(err));
   CfLog(cferror,OUTPUT,"");
   snprintf(hostbuffer,127,"(Non regitered IP)"); 
   return hostbuffer;
   }

for (ap = response; ap != NULL; ap = ap->ai_next)
   {   
   if ((err = getnameinfo(ap->ai_addr,ap->ai_addrlen,hostbuffer,127,0,0,0)) != 0)
      {
      snprintf(hostbuffer,127,"(Non registered IP)");
      freeaddrinfo(response);
      return hostbuffer;
      }
   
   Debug("Found address (%s) for host %s\n",hostbuffer,ipaddress);
   freeaddrinfo(response);
   return hostbuffer;
   }

 snprintf(hostbuffer,127,"(Non registered IP)"); 
#else
 snprintf(hostbuffer,127,"(Not for old BIND)"); 
#endif
 return hostbuffer;
}


/***************************************************************/

void LastSeen(hostname)

char *hostname;    

{ DBT key,value;
  DB *dbp;
  DB_ENV *dbenv = NULL;
  char name[bufsize];
  time_t now = time(NULL);

snprintf(name,bufsize-1,"%s/%s",VLOCKDIR,STATEDB_FILE);

if ((errno = db_create(&dbp,dbenv,0)) != 0)
   {
   snprintf(OUTPUT,bufsize*2,"Couldn't open last-seen database %s\n",name);
   CfLog(cferror,OUTPUT,"db_open");
   return;
   }

#ifdef CF_OLD_DB
if ((errno = dbp->open(dbp,name,NULL,DB_BTREE,DB_CREATE,0644)) != 0)
#else
if ((errno = dbp->open(dbp,NULL,name,NULL,DB_BTREE,DB_CREATE,0644)) != 0)
#endif
   {
   snprintf(OUTPUT,bufsize*2,"Couldn't open last-seen database %s\n",name);
   CfLog(cferror,OUTPUT,"db_open");
   dbp->close(dbp,0);
   return;
   }

bzero(&value,sizeof(value)); 
bzero(&key,sizeof(key));       
      
key.data = hostname;
key.size = strlen(hostname)+1;
value.data = (void *)&now;
value.size = sizeof(time_t); 

if ((errno = dbp->get(dbp,NULL,&key,&value,0)) == 0)
   {
   key.data = hostname;
   key.size = strlen(hostname)+1;
   value.data = (void *) &now;
   value.size = sizeof(time_t);
   
   Verbose("Updating last-seen time for %s\n",hostname);
   
   if ((errno = dbp->del(dbp,NULL,&key,0)) != 0)
      {
      CfLog(cferror,"","db->del");
      }
   
   key.data = hostname;
   key.size = strlen(hostname)+1;
   
   if ((errno = dbp->put(dbp,NULL,&key,&value,0)) != 0)
      {
      CfLog(cferror,"put failed","db->put");
      }
 
   dbp->close(dbp,0);
   }
else
   {
   key.data = hostname;
   key.size = strlen(hostname)+1;
   value.data = (void *) &now;
   value.size = sizeof(time_t);
   
   if ((errno = dbp->put(dbp,NULL,&key,&value,0)) != 0)
      {
      CfLog(cferror,"put failed","db->put");
      }
   
   dbp->close(dbp,0);
   }
}

/***************************************************************/

void CheckFriendConnections(hours)

/* Go through the database of recent connections and check for
   Long Time No See ...*/

int hours;

{ DBT key,value;
  DB *dbp;
  DBC *dbcp;
  DB_ENV *dbenv = NULL;
  int ret, secs = 3600*hours;
  struct stat statbuf;
  time_t now = time(NULL);
  char name[bufsize];
  
Verbose("CheckFriendConnections(%d)\n",hours);
snprintf(name,bufsize-1,"%s/%s",VLOCKDIR,STATEDB_FILE);

if ((errno = db_create(&dbp,dbenv,0)) != 0)
   {
   snprintf(OUTPUT,bufsize*2,"Couldn't open last-seen database %s\n",name);
   CfLog(cferror,OUTPUT,"db_open");
   return;
   }

#ifdef CF_OLD_DB
if ((errno = dbp->open(dbp,name,NULL,DB_BTREE,DB_CREATE,0644)) != 0)
#else
if ((errno = dbp->open(dbp,NULL,name,NULL,DB_BTREE,DB_CREATE,0644)) != 0)
#endif
   {
   snprintf(OUTPUT,bufsize*2,"Couldn't open last-seen database %s\n",name);
   CfLog(cferror,OUTPUT,"db_open");
   dbp->close(dbp,0);
   return;
   }

/* Acquire a cursor for the database. */

if ((ret = dbp->cursor(dbp, NULL, &dbcp, 0)) != 0)
   {
   CfLog(cferror,"Error reading from last-seen database","");
   dbp->err(dbp, ret, "DB->cursor");
   return;
   }

 /* Initialize the key/data return pair. */

memset(&key, 0, sizeof(key));
memset(&value, 0, sizeof(value));
 
 /* Walk through the database and print out the key/data pairs. */

while (dbcp->c_get(dbcp, &key, &value, DB_NEXT) == 0)
   {
   time_t then;

   bcopy(value.data,&then,sizeof(then));
   
   if (now > then + secs) 
      {
      snprintf(OUTPUT,bufsize,"Host %s last seen %s",(char *)key.data,ctime(&then));
      CfLog(cferror,OUTPUT,"");
      
      if (now > CFWEEK + then + 2*3600)
	 {
	 snprintf(OUTPUT,bufsize*2,"INFO: Giving up on %s, last seen more than a week ago at %s.",key.data,ctime(&then));
	 CfLog(cferror,OUTPUT,"");
	 
	 if ((errno = dbp->del(dbp,NULL,&key,0)) != 0)
	    {
	    CfLog(cferror,"","db_store");
	    }
	 }
      }
   else
      {      
      snprintf(OUTPUT,bufsize,"Host %s last seen %s",(char *)key.data,ctime(&then));
      CfLog(cfinform,OUTPUT,"");
      }
   }
 
dbcp->c_close(dbcp);
dbp->close(dbp,0);
}
